#include <string.h>
#include <assert.h>
#include "board.h"
#include "region.h"

#define BITCOUNT(x) \
	(((x) & 0x1) + \
	(((x) >> 1) & 0x1) + \
	(((x) >> 2) & 0x1) + \
	(((x) >> 3) & 0x1) + \
	(((x) >> 4) & 0x1) + \
	(((x) >> 5) & 0x1) + \
	(((x) >> 6) & 0x1) + \
	(((x) >> 7) & 0x1) + \
	(((x) >> 8) & 0x1))

static const Board g_zero;

void initialize_board(Board *b)
{
	*b = g_zero;
	generate_region(b, 0, 0);
}

void zero_board(Board *b)
{
	*b = g_zero;
}

void clear_number(Board *b)
{
	memset(b->x, 0, sizeof(b->x));
	memset(b->y, 0, sizeof(b->y));
	memset(b->b, 0, sizeof(b->b));
	memset(b->n, 0, sizeof(b->n));
	b->count = b->sol = 0;
}

int bit2int(unsigned int n)
{
	int i, v = 0;
	unsigned int b;
	
	for (i = 1, b = 0x1; i <= 9; i ++, b <<= 1) {
		v += (n & b) ? i : 0;
	}
	return v;
}

int count_possible_numbers(Board *b, int x, int y)
{
	int c = 0;
	unsigned int v = possible_number(b, x, y);
	
	while (v) {
		c += (v & 0x1);
		v >>= 1;
	}
	return c;
}

void put(Board *b, int x, int y, unsigned int v)
{
	assert(x >= 0 && x < 9 && y >= 0 && y < 9);
	assert(v == 0x1 || v == 0x2 || v == 0x4 ||
		   v == 0x8 || v == 0x10 || v == 0x11 ||
		   v == 0x8 || v == 0x10 || v == 0x20 ||
		   v == 0x40 || v == 0x80 || v == 0x100);
	b->n[y][x] = v;
	b->x[x] |= v;
	b->y[y] |= v;
	b->b[b->i[y][x]] |= v;
	b->count ++;
}

void rm(Board *b, int x, int y)
{
	unsigned int v = b->n[y][x];
	
	assert(x >= 0 && x < 9 && y >= 0 && y < 9);
	assert(v != 0);
	b->n[y][x] = 0;
	b->x[x] &= ~v;
	b->y[y] &= ~v;
	b->b[b->i[y][x]] &= ~v;
	b->count --;
}

int remove_number(Board *b, int x, int y)
{
	if (b->n[y][x]) {
		rm(b, x, y);
		return 0;
	}
	return 1;
}

int put_number(Board *b, int x, int y, int n)
{
	unsigned int v, m;

	if (n < 1 || n > 9) {
		remove_number(b, x, y);
		return 0;
	}
	m = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	v = 0x1 << (n - 1);
	if (m & v) {
		remove_number(b, x, y);
		put(b, x, y, v);
		return 0;
	}
	return 1;
}

void next(Board *b, int x, int y)
{
	unsigned int v, n;

	n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	v = b->n[y][x];
	if (v) {
		rm(b, x, y);
	}
	n &= ~v;
	if (v) n &= ~(v - 1);
	v = n & -n;
	if (v) {
		put(b, x, y, v);
	}
}

#if defined(_WIN32) && ! defined(__CONSOLE__)
int is_canceled(void);
#endif

#if 0

int search(Board *b, int x, int y)
{
	unsigned int n, m;
	int r;
	
	if (++ x == 9) {
#if defined(_WIN32) && ! defined(__CONSOLE__)
		if (is_canceled()) return -1;
#endif
		x = 0;
		if (++ y == 9) {
			return (++ b->sol > 1) ? 1 : 0;
		}
	}
	if (b->n[y][x]) {
		return search(b, x, y);
	}
	n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	while (n != 0) {
		m = n & -n;
		assert(m != 0);
		put(b, x, y, m);
		r = search(b, x, y);
		rm(b, x, y);
		if (r) return r;
		n &= ~m;
	}
	return 0;
}

#else

/*
 * return 
 * 1
 * < 0
 *
 */

static int search_(Board *, unsigned int o[9][9]);

static int decisive_fill(Board *b, unsigned int o[9][9])
{
	unsigned int n, m;
	int c, i, j, r, x, y, xx, yy;
	
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (b->n[y][x]) continue;
			n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
			c = 0;
			while (n != 0) {
				m = n & -n; n &= ~m; c ++;
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, x, y, m);
				r = search_(b, o);
				rm(b, x, y);
				return (r < 0) ? r : 1;
			}
		}
	}
	
	for (y = 0; y < 9; y ++) {
		n = ~(b->y[y]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (x = 0; x < 9; x ++) {
				if (! b->n[y][x] &&
				    (~(b->x[x] | b->b[b->i[y][x]]) & m) != 0) {
					c ++; xx = x;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, xx, y, m);
				r = search_(b, o);
				rm(b, xx, y);
				return (r < 0) ? r : 1;
			}
			n &= ~m;
		}
	}
	
	for (x = 0; x < 9; x ++) {
		n = ~(b->x[x]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (y = 0; y < 9; y ++) {
				if (! b->n[y][x] &&
				    (~(b->y[y] | b->b[b->i[y][x]]) & m) != 0) {
					c ++; yy = y;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, x, yy, m);
				r = search_(b, o);
				rm(b, x, yy);
				return (r < 0) ? r : 1;
			}
			n &= ~m;
		}
	}
	
	for (i = 0; i < 9; i ++) {
		n = ~(b->b[i]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (j = 0; j < 9; j ++) {
				x = b->j[i][j] >> 4;
				y = b->j[i][j] & 0xf;
				if (! b->n[y][x] &&
				    (~(b->y[y] | b->x[x]) & m) != 0) {
					c ++; xx = x; yy = y;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, xx, yy, m);
				r = search_(b, o);
				rm(b, xx, yy);
				return (r < 0) ? r : 1;
			}
			n &= ~m;
		}
	}
	return 0;
}

int search_minimum_choice(Board *b, int *x, int *y)
{
	int i, j, n, m, r;
	
	m = 10; r = 0;
	for (i = 0; i < 9; i ++) {
		for (j = 0; j < 9; j ++) {
			if (b->n[i][j]) continue;
			n = ~(b->x[j] | b->y[i] | b->b[b->i[i][j]]) & 0x1ff;
			r = 0;
			while (n != 0) {
				n &= ~(n & -n); r ++;
			}
			assert(r > 0 && r <= 9);
			if (r <= m) {
				*x = j; *y = i; m = r;
			}
		}
	}
	return m % 10;
}

static int search_(Board *b, unsigned int o[9][9])
{
	unsigned int n, m;
	int x, y, r;
	
#if defined(_WIN32) && ! defined(__CONSOLE__)
	if (is_canceled()) return -2;
#endif
	r = decisive_fill(b, o);
	if (r < 0) return r;
	if (r != 0) return 0;
	if (! search_minimum_choice(b, &x, &y)) {
		if (b->sol > 0) {
			if (memcmp(b->n, o, sizeof(b->n))) {
				b->sol ++;
				return -1;
			}
		} else {
			b->sol ++;
			memcpy(o, b->n, sizeof(b->n));
		}
		return 0;
	}
	n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	while (n != 0) {
		m = n & -n;
		assert(m != 0);
		put(b, x, y, m);
		r = search_(b, o);
		rm(b, x, y);
		if (r) return r;
		n &= ~m;
	}
	return 0;
}

int search(Board *b, int x, int y)
{
	unsigned int o[9][9];
	
	return search_(b, o);
}
#endif
