#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <curses.h>
#include <getopt.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/types.h>

#include "board.h"
#include "gen.h"
#include "strw.h"
#include "randomh"
#include "region.h"

#define LEFT 8
#define TOP 2

#define SAVEFILE ".npw"

#define QUIT_KEY 'q'
#define REDRAW_KEY 'r'
#define CONFIRM_LOAD          "load? (y/n)"
#define CONFIRM_SAVE          "save? (y/n)"
#define CONFIRM_DELETE        "delete? (y/n)"
#define CONFIRM_RESTART       "restart? (y/n)"
#define CONFIRM_QUIT          "quit? (y/n)"
#define CONFIRM_CLEAR_MESSAGE "              "

const char *get_home_dir(void)
{
	const char *home;
	struct passwd *pw;
	home = getenv("HOME");
	if (home) {
		return home;
	}
	pw = getpwuid(getuid());
	if (pw) {
		return pw->pw_dir;
	}
	return NULL;
}

int load_stage(Board *b, Board *p)
{
	const char *home, *file;
	char *path;
	char code_b[B_CODELEN + 1];
	char code_r[B_CODELEN + 1];
	size_t size;
	FILE *fp;
	int result = 0;
	
	home = get_home_dir();
	size = strlen(home) + sizeof(SAVEFILE) + 1;
	path = malloc(size);
	if (path) {
		sprintf(path, "%s/%s", home, SAVEFILE);
		file = path;
	} else {
		file = SAVEFILE;
	}
	do {
		fp = fopen(file, "r");
		free(path);
		if (fp) {
			fgets(code_b, sizeof(code_b), fp);
			fgets(code_r, sizeof(code_r), fp);
			fclose(fp);
			if (! code2region(p, code_r)) return 0;
			*b = *p;
			if (! code2board(p, b, code_b)) return 0;
			result = 1;
		} else if (path) {
			path = NULL;
			file = SAVEFILE;
			continue;
		}
		break;
	} while (1);
	return result;
}

void save_stage(Board *b, Board *p)
{
	const char *home, *file;
	char *path;
	char code[B_CODELEN];
	size_t size;
	FILE *fp;
	
	home = get_home_dir();
	size = strlen(home) + sizeof(SAVEFILE) + 1;
	path = malloc(size);
	if (path) {
		sprintf(path, "%s/%s", home, SAVEFILE);
		file = path;
	} else {
		file = SAVEFILE;
	}
	fp = fopen(file, "w");
	if (fp) {
		board2code(p, b, code);
		fputs(code, fp);
		fputc('\n', fp);
		region2code(p, code);
		fputs(code, fp);
		fputc('\n', fp);
		fclose(fp);
	}
	free(path);
}

void delete_stage(void)
{
	const char *home, *file;
	char *path;
	size_t size;
	
	home = get_home_dir();
	size = strlen(home) + sizeof(SAVEFILE) + 1;
	path = malloc(size);
	if (path) {
		sprintf(path, "%s/%s", home, SAVEFILE);
		file = path;
	} else {
		file = SAVEFILE;
	}
	unlink(file);
	free(path);
}

void generate_stage(Board *b, Board *p)
{
	change_seed();
	clear();
	move(TOP + 13, LEFT);
	addstr("generating");
	refresh();
	initialize_board(b);
	generate_region(b, 1, 0);
	generate(b, -1, 0);
	attrset(A_NORMAL);
	random_scan_rm(b);
	*p = *b;
}

void display(Board *b, Board *p)
{
	const static char *right[] = {
		"+-------------------+",
		"|   number place    |",
		"+-------------------+",
		"| left, h, ^b       |",
		"|       cursor left |",
		"| right, l, ^f      |",
		"|      cursor right |",
		"| up, k, ^p         |",
		"|         cursor up |",
		"| down, j, ^n       |",
		"|       cursor down |",
		"| 0123456789        |",
		"|       put number  |",
		"| enter, space      |",
		"| ^m, n, x          |",
		"|        put number |",
		"| delete, backspace |",
		"| ^h, b, z          |",
		"|     remove number |",
		"+-------------------+",
		NULL,
	};
	static const char n[] = {
		'.', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	};
	static const char w[] = {
		' ', '+', '+', '+', '+', '+', '-', '+',
		'+', '|', '+', '+', '+', '+', '+', '+'
	};
	int i, x, y, v;
	char num[13] = {0};
	
	attrset(A_NORMAL);
	/* */
	snprintf(num, sizeof(num) - 1, "numbers: %2d", b->count);
	mvaddstr(TOP, LEFT, num);
	/* */
	move(TOP, LEFT);
	addstr("+---");
	for (x = 0; x < 8; x ++) {
		addch((b->i[0][x] == b->i[0][x + 1]) ? '-' : '+');
		addstr("---"); 
	}
	addstr("+"); 
	/* */
	for (y = 0; y < 9; y ++) {
		move(TOP + 2 * y + 1, LEFT);
		addch('|');
		for (x = 0; x < 9; x ++) {
			addch(' ');
			if (p->n[y][x]) {
				attron(A_STANDOUT);
				addch(n[bit2int(b->n[y][x])]);
				attroff(A_STANDOUT);
			} else {
				addch(n[bit2int(b->n[y][x])]);
			}
			addch(' ');
			if (x < 8 && b->i[y][x] == b->i[y][x + 1]) {
				addch(' ');
			} else {
				addch('|');
			}
		}
		move(TOP + 2 * y + 2, LEFT);
		addch((y < 8 && b->i[y][0] == b->i[y + 1][0]) ? '|' : '+');
		for (x = 0; x < 8; x ++) {
			if (y < 8 && b->i[y][x] == b->i[y + 1][x]) {
				addstr("   ");
			} else {
				addstr("---");
			}
			v = 0;
			v |= (b->i[y][x] != b->i[y][x + 1]) ? 0x1 : 0;
			if (y < 8) {
				v |= (b->i[y][x] != b->i[y + 1][x]) ? 0x2 : 0;
				v |= (b->i[y][x + 1] != b->i[y + 1][x + 1]) ? 0x4 : 0;
				v |= (b->i[y + 1][x] != b->i[y + 1][x + 1]) ? 0x8 : 0;
			} else {
				v |= 0x6;
			}
			addch(w[v]);
		}
		if (y < 8 && b->i[y][x] == b->i[y + 1][x]) {
			addstr("   |");
		} else {
			addstr("---+");
		}
	}
	/* */
	for (i = 0; right[i] != NULL; i ++) {
		mvaddstr(TOP + i, LEFT + 40, right[i]);
	}
}

int main_loop(Board *b, Board *p)
{
	int c, x, y, xx, yy;

	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (! p->n[y][x]) {
				goto START;
			}
		}
	}
	return 0;
START:
	while (1) {
		display(b, p);
		xx = LEFT + x * 4 + 2;
		yy = TOP + y * 2 + 1;
		move(yy, xx);
		refresh();
		if (b->count == 9 * 9) {
			break;
		}
		c = getch();
#ifdef _DEBUG
		{
			char p[33]; sprintf(p, "%5o", c);
			mvaddstr(0, 0, p);
			mvaddch(1, 0, c);
		}
#endif
		switch (c) {
		case KEY_DOWN:
		case 016: /* ^N */
		case 'j': case 'J':
			y ++;
			if (y >= 9) {
				y = 0;
			}
			break;
		case KEY_UP:
		case 020: /* ^N */
		case 'k': case 'K':
			y --;
			if (y < 0) {
				y = 8;
			}
			break;
		case KEY_LEFT:
		case 02: /* ^B */
		case 'h': case 'H':
			x --;
			if (x < 0) {
				x = 8;
			}
			break;
		case KEY_RIGHT:
		case 06: /* ^F */
		case 'l': case 'L':
			x ++;
			if (x >= 9) {
				x = 0;
			}
			break;
		case KEY_BACKSPACE:
		case KEY_DC:
		case 010: /* ^H */
		case 'd': case 'D':
		case 'z': case 'Z':
		case 'b': case 'B':
			if (! p->n[y][x] && b->n[y][x]) {
				rm(b, x, y);
			}
			break;
		case KEY_ENTER:
		case 012: /* ^M */
		case ' ':
		case 'x': case 'X':
		case 'n': case 'N':
			if (! p->n[y][x]) {
				next(b, x, y);
			}
			break;
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			if (! p->n[y][x]) {
				put_number(b, x, y, c - '0');
			}
			break;
		case 'q': case 'Q':
			return 1;
		}
	}
	return 0;
}

int confirm(const char *message)
{
	int c;
	move(TOP + 21, LEFT);
	addstr(message);
	do {
		c = getch();
	} while (c != 'y' && c != 'Y' && c != 'n' && c != 'N');
	move(TOP + 21, LEFT);
	addstr(CONFIRM_CLEAR_MESSAGE);
	return c == 'y' || c == 'Y';
}

void onexit(void)
{
	endwin();
}

int main(void)
{
	Board b, b2;
	time_t sec;
	char s[80];
	
	initscr();
	atexit(onexit);
	noecho();
	cbreak();
	keypad(stdscr, TRUE);
	init_rand();
	if (load_stage(&b, &b2)) {
		display(&b, &b2);
		if (! confirm(CONFIRM_LOAD)) {
			if (confirm(CONFIRM_DELETE)) {
				delete_stage();
			}
			generate_stage(&b, &b2);
		}
	} else {
		generate_stage(&b, &b2);
	}
	do {
		sec = time(NULL);
		do {
			if (! main_loop(&b, &b2)) {
				time_t min;
				sec = difftime(time(NULL), sec);
				min = sec / 60;
				sec %= 60;
				snprintf(s, sizeof(s), "%lu:%02lu", min, sec);
				move(TOP + 20, LEFT);
				addstr(s);
				break;
			}
			if (confirm(CONFIRM_SAVE)) {
				save_stage(&b, &b2);
				break;
			}
			if (! confirm(CONFIRM_RESTART)) break;
			b = b2;
		} while (1);
		if (confirm(CONFIRM_QUIT)) break;
		generate_stage(&b, &b2);
	} while (1);
	
	nocbreak();
	echo();
	refresh();
	
	return 0;
}
