#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <process.h>
#include <tchar.h>
#include <assert.h>

#include "Resnpw.h"

#include "board.h"
#include "symmetry.h"
#include "gen.h"
#include "region.h"
#include "strw.h"
#include "util.h"
#include "mark.h"
#include "aa.h"
#include "random.h"

#define ID_TOOLBAR  1001
#define ID_STATUS   1002

typedef enum {EDIT, GENERATE, SOLVE} State;
typedef enum {NEW, FIX, RETRY, RELEASE, CLEAR} Manipulate;
typedef enum {SYMMETRY_NONE, SYMMETRY_POLAR, SYMMETRY_RANDOM} SymmetryType;

static const TCHAR g_lpszApplicationName[] = "npw";
static const TCHAR g_lpszProfileFileName[] = "npw.ini";

static const TCHAR g_lpszSection[] = "default";
static const TCHAR g_lpszColorSection[] = "color";

static HINSTANCE g_hInstance;
static HWND g_hwndFrame, g_hwndToolbar, g_hwndStatus;
static HACCEL g_hAccel;

static HMENU g_hmenuContext, g_hmenuContextMark;
static HFONT g_hFont, g_hSmallFont;
static HPEN g_hPen, g_hBoldPen;
static HBRUSH g_hRegionBrush[9];

static LONG g_matrix_size, g_grid_size;
static POINT g_origin;

static Board g_board, g_playing_board;
static State g_state;
static TCHAR g_lpszFileName[MAX_PATH];

/* view parameter */
static BOOL g_hint, g_paint_region, g_paint_axis;
/* edit parameter */
static BOOL g_mark_mode = FALSE, g_mark_exclusion = FALSE;
static unsigned int g_mark_mask = 0x1ff;
/* configurable parameter */
static SymmetryType g_symmetry = SYMMETRY_POLAR;
static BOOL g_autosave = FALSE;
static BOOL g_region = FALSE;
static BOOL g_mark_type = FALSE;
static int g_cursor_pos = 0;
/* color */
static COLORREF g_region_color[9] = {
	RGB(0xff, 0xff, 0xff), RGB(0xcc, 0xff, 0xff), RGB(0xff, 0xcc, 0xff),
	RGB(0xff, 0xff, 0xcc), RGB(0xff, 0xcc, 0xcc), RGB(0xcc, 0xff, 0xcc),
	RGB(0xcc, 0xcc, 0xff), RGB(0xff, 0xff, 0x66), RGB(0x66, 0xff, 0xff),
};
static COLORREF g_number_color = RGB(0, 0, 0);
static COLORREF g_fix_color = RGB(0xff, 0, 0);
static COLORREF g_mark_color = RGB(0, 0, 0xff);

static void release_resource(void)
{
	int i;
	
	if (g_hmenuContext) DestroyMenu(g_hmenuContext);
	if (g_hmenuContextMark) DestroyMenu(g_hmenuContextMark);
	if (g_hFont) DeleteObject(g_hFont);
	if (g_hSmallFont) DeleteObject(g_hSmallFont);
	if (g_hPen) DeleteObject(g_hPen);
	if (g_hBoldPen) DeleteObject(g_hBoldPen);
	for (i = 0; i < 9; i ++) {
		if (g_hRegionBrush[i]) DeleteObject(g_hRegionBrush[i]);
	}
}

static void create_region_brush(void)
{
	int i;
	
	for (i = 0; i < 9; i ++) {
		if (g_hRegionBrush[i]) DeleteObject(g_hRegionBrush[i]);
		g_hRegionBrush[i] = CreateSolidBrush(g_region_color[i]);
	}
}

BOOL CALLBACK WaitDialogProc(HWND hwnd, UINT uMsg,
                             WPARAM wParam, LPARAM lParam)
{
	unsigned long hThread;
	
	switch (uMsg) {
		case WM_INITDIALOG:
			center_on_parent_window(hwnd);
			SetWindowLongPtr(hwnd, DWL_USER, (LONG) ((void **) lParam)[1]);
			hThread = _beginthread(
				(void (*)(void *)) ((void **) lParam)[0], 0, hwnd);
			if (hThread == -1) {
				EndDialog(hwnd, FALSE);
			}
			break;
		case WM_COMMAND:
			g_state = EDIT;
			if (GetWindowLongPtr(hwnd, DWL_USER)) {
				EndDialog(hwnd, FALSE);
			}
			break;
		default:
			return FALSE;
	}
	return TRUE;
}


BOOL color_config(LPCTSTR lpPath, BOOL is_save)
{
	TCHAR key[sizeof("region") + 1];
	TCHAR lpValue[33];
	int i;
	BOOL r = TRUE;
	
	for (i = 0; i < 9; i ++) {
		wsprintf(key, "region%d", i + 1);
		if (is_save) {
			wsprintf(lpValue, "0x%06x", g_region_color[i]);
			r &= WritePrivateProfileString(g_lpszColorSection, key, lpValue, lpPath);
		} else {
			g_region_color[i] = GetPrivateProfileInt(g_lpszColorSection, key, g_region_color[i], lpPath);
		}
	}
	if (is_save) {
		wsprintf(lpValue, "0x%06x", g_number_color);
		r &= WritePrivateProfileString(g_lpszColorSection, "number", lpValue, lpPath);
		wsprintf(lpValue, "0x%06x", g_fix_color);
		r &= WritePrivateProfileString(g_lpszColorSection, "fix", lpValue, lpPath);
		wsprintf(lpValue, "0x%06x", g_mark_color);
		r &= WritePrivateProfileString(g_lpszColorSection, "mark", lpValue, lpPath);
	} else {
		g_number_color = GetPrivateProfileInt(g_lpszColorSection, "number", g_number_color, lpPath);
		g_fix_color = GetPrivateProfileInt(g_lpszColorSection, "fix", g_fix_color, lpPath);
		g_mark_color = GetPrivateProfileInt(g_lpszColorSection, "mark", g_mark_color, lpPath);
	}
	return r;
}

void load_config(LPINT x, LPINT y, LPINT cx, LPINT cy)
{
	LPTSTR lpPath;
	HWND hwndDesktop;
	RECT rc;
	
	lpPath = get_profile_path(g_lpszProfileFileName);
	if (! lpPath) {
		*x = *y = *cx = *cy = CW_USEDEFAULT;
		return;
	}
	
	*x = GetPrivateProfileInt(g_lpszSection, "x", CW_USEDEFAULT, lpPath);
	*y = GetPrivateProfileInt(g_lpszSection, "y", CW_USEDEFAULT, lpPath);
	*cx = GetPrivateProfileInt(g_lpszSection, "cx", CW_USEDEFAULT, lpPath);
	*cy = GetPrivateProfileInt(g_lpszSection, "cy", CW_USEDEFAULT, lpPath);
	
	g_symmetry = GetPrivateProfileInt(g_lpszSection, "symmetry", SYMMETRY_POLAR, lpPath);
	if (g_symmetry < SYMMETRY_NONE || g_symmetry > SYMMETRY_RANDOM) {
		g_symmetry = SYMMETRY_POLAR;
	}
	
	g_autosave = GetPrivateProfileInt(g_lpszSection, "autosave", g_autosave, lpPath);
	g_region = GetPrivateProfileInt(g_lpszSection, "jigsaw", g_region, lpPath);
	g_mark_type = GetPrivateProfileInt(g_lpszSection, "mark", g_mark_type, lpPath);
	g_cursor_pos = GetPrivateProfileInt(g_lpszSection, "pos", g_cursor_pos, lpPath);
	
	color_config(lpPath, FALSE);
	
	free(lpPath);
	
	hwndDesktop = GetDesktopWindow();
	if (GetWindowRect(hwndDesktop, &rc)) {
		if (*x < rc.left || *y < rc.top ||
		    *x + *cx > rc.right - rc.left ||
		    *y + *cy > rc.bottom - rc.top) {
			*x = *y = *cx = *cy = CW_USEDEFAULT;
		}
	}
	return;
}

BOOL save_config(HWND hwnd)
{
	RECT rc;
	INT x, y, cx, cy;
	TCHAR lpValue[33];
	LPTSTR lpPath;
	BOOL r = TRUE;
	
	lpPath = get_profile_path(g_lpszProfileFileName);
	if (! lpPath) return FALSE;
	
	if (! IsIconic(hwnd) && ! IsZoomed(hwnd)) {
		GetWindowRect(hwnd, &rc);
		x = rc.left;
		y = rc.top;
		cx = rc.right - rc.left;
		cy = rc.bottom - rc.top;
		wsprintf(lpValue, "%d", x);
		r &= WritePrivateProfileString(g_lpszSection, "x", lpValue, lpPath);
		wsprintf(lpValue, "%d", y);
		r &= WritePrivateProfileString(g_lpszSection, "y", lpValue, lpPath);
		wsprintf(lpValue, "%d", cx);
		r &= WritePrivateProfileString(g_lpszSection, "cx", lpValue, lpPath);
		wsprintf(lpValue, "%d", cy);
		r &= WritePrivateProfileString(g_lpszSection, "cy", lpValue, lpPath);
	}
	
	wsprintf(lpValue, "%d", g_symmetry);
	r &= WritePrivateProfileString(g_lpszSection, "symmetry", lpValue, lpPath);
	wsprintf(lpValue, "%d", g_autosave);
	r &= WritePrivateProfileString(g_lpszSection, "autosave", lpValue, lpPath);
	wsprintf(lpValue, "%d", g_region);
	r &= WritePrivateProfileString(g_lpszSection, "jigsaw", lpValue, lpPath);
	wsprintf(lpValue, "%d", g_mark_type);
	r &= WritePrivateProfileString(g_lpszSection, "mark", lpValue, lpPath);
	wsprintf(lpValue, "%d", g_cursor_pos);
	r &= WritePrivateProfileString(g_lpszSection, "pos", lpValue, lpPath);
	
	r &= color_config(lpPath, TRUE);
	
	free(lpPath);
	
	return r;
}


void redraw(HWND hwnd, RECT *lprc)
{
	if (! lprc) {
		RECT rc[3] = {{0}, {0}, {0}};
		
		if (GetClientRect(hwnd, &rc[0]) &&
		    GetWindowRect(g_hwndToolbar, &rc[1]) &&
		    GetWindowRect(g_hwndStatus, &rc[2])) {
			rc[0].top += rc[1].bottom - rc[1].top;
			rc[0].bottom -= rc[2].bottom - rc[2].top;
			InvalidateRect(hwnd, &rc[0], FALSE);
			UpdateWindow(hwnd);
		}
	} else {
		InvalidateRect(hwnd, lprc, FALSE);
		UpdateWindow(hwnd);
	}
}


void update_status(void)
{
	set_statusbar_text_id(g_hInstance, g_hwndStatus,
		(g_mark_mode) ? IDS_MARK : IDS_EDITING,
		g_board.count, g_playing_board.count);
}

void update_menu_item(HMENU hmenu, const UINT item, BOOL pref)
{
	CheckMenuItem(hmenu, item,
	              MF_BYCOMMAND | ((pref) ? MF_CHECKED : MF_UNCHECKED));
}

void update_copy_menu(HMENU hmenu, BOOL flag)
{
	UINT items[] = {
		IDM_COPY,
		IDM_COPY_SPC,
		IDM_COPY_ZUNTAN,
		IDM_COPY_ZUNTAN_URL, 
		IDM_COPY_BRAINBASHERS,
		IDM_COPY_PLAYR_CO_UK,
		IDM_COPY_CSV,
		IDM_COPY_AXIS,
		IDM_COPY_AXIS_FIXED,
		IDM_COPY_PROBLEM,
		IDM_COPY_SPC_PROBLEM,
		IDM_COPY_ZUNTAN_PROBLEM,
		IDM_COPY_ZUNTAN_URL_PROBLEM,
		IDM_COPY_BRAINBASHERS_PROBLEM,
		IDM_COPY_PLAYR_CO_UK_PROBLEM,
		IDM_COPY_CSV_PROBLEM,
		IDM_COPY_AXIS_PROBLEM,
		IDM_COPY_AXIS_FIXED_PROBLEM,
	};
	UINT fuFlags;
	int i;
	
	fuFlags = MF_BYCOMMAND | (flag ? MF_ENABLED : MF_GRAYED);
	for (i = 0; i < sizeof(items) / sizeof(items[0]); i ++) {
		EnableMenuItem(hmenu, items[i], fuFlags);
	}
}

void update_context_menu(HMENU hmenu, int x, int y)
{
	int i;
	unsigned int n;
	
	/* manipulate mark item */
	n = g_mark[y][x] & 0x1ff;
	for (i = 1; i <= 9; i ++) {
		CheckMenuItem(hmenu, i + 12, (n & 0x1) ? MF_CHECKED : MF_UNCHECKED);
		n >>= 1;
	}
	EnableMenuItem(hmenu, 12, (g_mark[y][x]) ? MF_ENABLED : MF_GRAYED);
	/* manipulate put item */
	n = ~(g_playing_board.x[x] | g_playing_board.y[y] | g_playing_board.b[g_playing_board.i[y][x]]) & 0x1ff;
	for (i = 1; i <= 9; i ++) {
		EnableMenuItem(hmenu, i + 1, (n & 0x1) ? MF_ENABLED : MF_GRAYED);
		n >>= 1;
	}
	EnableMenuItem(hmenu, 1, (g_playing_board.n[y][x]) ? MF_ENABLED : MF_GRAYED);
}

void update_symmetry_menu(HMENU hmenu, unsigned int flag)
{
	switch (g_symmetry) {
	case SYMMETRY_NONE:
		CheckMenuItem(hmenu, IDM_NO_SYMMETRY, MF_BYCOMMAND | flag);
		break;
	case SYMMETRY_RANDOM:
		CheckMenuItem(hmenu, IDM_SYMMETRY, MF_BYCOMMAND | flag);
		break;
	case SYMMETRY_POLAR:
		CheckMenuItem(hmenu, IDM_SYMMETRY_POLAR, MF_BYCOMMAND | flag);
		break;
	}
}

void change_symmetry_setting(HWND hwnd, SymmetryType st)
{
	unsigned int f = MF_UNCHECKED;
	
	if (g_symmetry == st) return;
	for (;;) {
		update_symmetry_menu(GetMenu(hwnd), f);
		if (f == MF_CHECKED) return;
		f = MF_CHECKED;
		g_symmetry = st;
	}
}

void change_mark_shown(HWND hwnd, int n)
{
	static const UINT id[] = {
		IDM_HINT_1, IDM_HINT_2, IDM_HINT_3, IDM_HINT_4, IDM_HINT_5,
		IDM_HINT_6, IDM_HINT_7, IDM_HINT_8, IDM_HINT_9,
	};
	int i;
	unsigned int v;
	
	if (n >= 1 && n <= 9) {
		v = 0x1 << (n - 1);
		g_mark_mask ^= v;
		update_menu_item(GetMenu(hwnd), id[n - 1], g_mark_mask & v);
		SendMessage(g_hwndToolbar, TB_CHECKBUTTON, id[n - 1],
		            MAKELONG(g_mark_mask & v, 0));
		redraw(hwnd, NULL);
		return;
	}
	switch (n) {
	case 10:
		g_mark_mask = 0x1ff;
		break;
	case 11:
		g_mark_mask = 0;
		break;
	}
	for (i = 0, v = 1; i < 9; i ++, v <<= 1) {
		update_menu_item(GetMenu(hwnd), id[i], g_mark_mask & v);
		SendMessage(g_hwndToolbar, TB_CHECKBUTTON, id[i],
		            MAKELONG(g_mark_mask & v, 0));
	}
	if (n) {
		redraw(hwnd, NULL);
	}
}

BOOL load_state(LPCTSTR lpPath)
{
	TCHAR lpState[B_CODELEN] = {0};
	TCHAR lpMark[B_MARKCODELEN] = {0};
	Board b, p;
	
	/* region */
	zero_board(&b);
	GetPrivateProfileString(g_lpszSection, "region", "", lpState, B_CODELEN, lpPath);
	if (lpState[0] != 0) {
		if (! code2region(&b, lpState) || check_region(&b)) return FALSE;
	} else {
		initialize_board(&b);
	}
	p = b;
	/* board */
	lpState[0] = _T('\0');
	GetPrivateProfileString(g_lpszSection, "state", "", lpState, B_CODELEN, lpPath);
	if (lpState[0] == 0 || ! code2board(&b, &p, lpState)) {
		return FALSE;
	}
	g_board = b;
	g_playing_board = p;
	/* mark */
	GetPrivateProfileString(g_lpszSection, "mark", "", lpMark, B_MARKCODELEN, lpPath);
	if (lpMark[0] == 0 || ! code2mark(lpMark)) {
		mark_clear_all();
	}
	/* etc */
	g_paint_region = GetPrivateProfileInt(g_lpszSection, "region_color", g_paint_region, lpPath);
	g_mark_exclusion = GetPrivateProfileInt(g_lpszSection, "mark_exclusion", g_mark_exclusion, lpPath);
	g_paint_axis = GetPrivateProfileInt(g_lpszSection, "axis", g_paint_axis, lpPath);
	
	return TRUE;
}

BOOL save_state(LPCTSTR lpPath)
{
	TCHAR lpState[B_CODELEN] = {0};
	TCHAR lpMark[B_MARKCODELEN] = {0};
	TCHAR lpValue[33];
	BOOL r = TRUE;
	
	region2code(&g_board, lpState);
	assert(lstrlen(lpState) < B_CODELEN);
	r &= WritePrivateProfileString(g_lpszSection, "region", lpState, lpPath);
	
	board2code(&g_board, &g_playing_board, lpState);
	assert(lstrlen(lpState) < B_CODELEN);
	r &= WritePrivateProfileString(g_lpszSection, "state", lpState, lpPath);
	
	mark2code(lpMark);
	assert(lstrlen(lpMark) < B_MARKCODELEN);
	r &= WritePrivateProfileString(g_lpszSection, "mark", lpMark, lpPath);
	
	wsprintf(lpValue, "%d", g_paint_region);
	r &= WritePrivateProfileString(g_lpszSection, "region_color", lpValue, lpPath);
	
	wsprintf(lpValue, "%d", g_mark_exclusion);
	r &= WritePrivateProfileString(g_lpszSection, "mark_exclusion", lpValue, lpPath);
	
	wsprintf(lpValue, "%d", g_paint_axis);
	r &= WritePrivateProfileString(g_lpszSection, "axis", lpValue, lpPath);
	
	return r;
}

void set_default_filename(HWND hwnd)
{
	LPTSTR lpPath;
	
	lpPath = get_profile_path(g_lpszProfileFileName);
	if (lpPath) {
		lstrcpy(g_lpszFileName, lpPath);
		free(lpPath);
	} else {
		lstrcpy(g_lpszFileName, g_lpszProfileFileName);
	}
	if (hwnd) {
		set_window_text(hwnd, g_lpszApplicationName);
	}
}

void on_open(HWND hwnd, LPTSTR path)
{
	LPTSTR lpPath;
	BOOL result;
	
	lpPath = (path) ? path : g_lpszFileName;
	result = load_state(lpPath);
	if (result) {
		update_copy_menu(GetMenu(hwnd), ! is_jigsaw(&g_board));
		update_menu_item(GetMenu(hwnd), IDM_COLOR, g_paint_region);
		update_menu_item(GetMenu(hwnd), IDM_MARK_EXCLUSION, g_mark_exclusion);
		update_menu_item(GetMenu(hwnd), IDM_AXIS, g_paint_axis);
		if (path) {
			lstrcpyn(g_lpszFileName, path, MAX_PATH);
		}
		set_window_text(hwnd, "%s - %s", g_lpszApplicationName, get_basename(lpPath));
		update_status();
		redraw(hwnd, NULL);
	} else {
		message_box_id(g_hInstance, hwnd, IDS_LOAD_ERROR, g_lpszApplicationName, MB_ICONASTERISK | MB_OK);
	}
}

void on_save(HWND hwnd, LPTSTR path)
{
	LPTSTR lpPath;
	BOOL result;
	
	lpPath = (path) ? path : g_lpszFileName;
	result = save_state(lpPath);
	if (result) {
		if (path) {
			lstrcpyn(g_lpszFileName, path, MAX_PATH);
		}
		set_window_text(hwnd, "%s - %s", g_lpszApplicationName, get_basename(lpPath));
	} else {
		message_box_id(g_hInstance, hwnd, IDS_SAVE_ERROR, g_lpszApplicationName, MB_ICONASTERISK | MB_OK);
	}
}

void calc_metrics_size(LPRECT rc, LPLONG lpSize, LPLONG lpGrid,
	LPPOINT lpOrigin)
{
	LONG width, height, size, grid;
	
	width = rc->right - rc->left;
	height = rc->bottom - rc->top;
	if (width > height) {
		size = height;
	} else {
		size = width;
	}
	grid = size / 10;
	size = grid * 9;
	lpOrigin->x = (width - size) / 2 + rc->left;
	lpOrigin->y = (height - size) / 2 + rc->top;
	*lpGrid = grid;
	*lpSize = size;
}

void update_paint_param(LPRECT rc)
{
	LOGFONT logfont = {0};
	LONG size, grid;
	POINT org;
	
	calc_metrics_size(rc, &size, &grid, &org);
	g_origin = org;
	if (grid == g_grid_size) return;
	g_matrix_size = size;
	g_grid_size = grid;
	
	if (g_hFont) {
		DeleteObject(g_hFont);
	}
	size = 4 * grid / 5;
	if (size == 0) size = 1;
	logfont.lfHeight = size;
	size = 4 * grid / 7;
	if (size == 0) size = 1;
	logfont.lfWidth = size;
#if defined(WINVER) && WINVER >= 0x0400
	logfont.lfQuality = ANTIALIASED_QUALITY;
#else
	logfont.lfQuality = DEFAULT_QUALITY;
#endif
	lstrcpy(logfont.lfFaceName, "Arial");
	g_hFont = CreateFontIndirect(&logfont);
	
	if (g_hSmallFont) {
		DeleteObject(g_hSmallFont);
	}
	size = 4 * grid / 15;
	if (size == 0) size = 1;
	logfont.lfHeight = size;
	size = 4 * grid / 21;
	if (size == 0) size = 1;
	logfont.lfWidth = size;
	g_hSmallFont = CreateFontIndirect(&logfont);
	
	if (g_hPen) {
		DeleteObject(g_hPen);
	}
	g_hPen = CreatePen(PS_SOLID, grid / 80 + 1, RGB(0, 0, 0));
	
	if (g_hBoldPen) {
		DeleteObject(g_hBoldPen);
	}
	g_hBoldPen = CreatePen(PS_SOLID, grid / 20 + 1, RGB(0, 0, 0));
}

void screen_to_matrix(int *x, int *y)
{
	if (*x < g_origin.x || *x > g_origin.x + g_matrix_size) {
		*x = -1;
	} else {
		*x = (*x - g_origin.x) / g_grid_size;
	}
	if (*y < g_origin.y || *y > g_origin.y + g_matrix_size) {
		*y = -1;
	} else {
		*y = (*y - g_origin.y) / g_grid_size;
	}
}

int screen_to_matrix_cell(LPLONG x, LPLONG y)
{
	static const int s = 6;
	int out = 0, dx, dy;
	
	switch (g_cursor_pos) {
	case 0:
		dx = g_grid_size / s;
		dy = (s - 1) * g_grid_size / s;
		break;
	case 1:
		dx = g_grid_size / 2;
		dy = (s - 1) * g_grid_size / s;
		break;
	case 2:
		dx = (s - 1) * g_grid_size / s;
		dy = (s - 1) * g_grid_size / s;
		break;
	case 3:
		dx = g_grid_size / s;
		dy = g_grid_size / 2;
		break;
	case 4:
		dx = g_grid_size / 2;
		dy = g_grid_size / 2;
		break;
	case 5:
		dx = (s - 1) * g_grid_size / s;
		dy = g_grid_size / 2;
		break;
	case 6:
		dx = g_grid_size / s;
		dy = g_grid_size / s;
		break;
	case 7:
		dx = g_grid_size / 2;
		dy = g_grid_size / s;
		break;
	case 8:
		dx = (s - 1) * g_grid_size / s;
		dy = g_grid_size / s;
		break;
	default:
		g_cursor_pos = 4;
		dx = g_grid_size / 2;
		dy = g_grid_size / 2;
		break;
	}
	if (*x < g_origin.x) {
		*x = g_origin.x + dx;
		out = 1;
	} else if (*x > g_origin.x + g_matrix_size) {
		*x = g_origin.x + g_matrix_size - g_grid_size + dx;
		out = 1;
	} else {
		*x = g_origin.x + (*x - g_origin.x) / g_grid_size * g_grid_size + dx;
	}
	if (*y < g_origin.y) {
		*y = g_origin.y + dy;
		out = 1;
	} else if (*y > g_origin.y + g_matrix_size) {
		*y = g_origin.y + g_matrix_size - g_grid_size + dy;
		out = 1;
	} else {
		*y = g_origin.y + (*y - g_origin.y) / g_grid_size * g_grid_size + dy;
	}
	return out;
}

void update_matrix(HWND hwnd, LONG x, LONG y, BOOL number)
{
	RECT rc;
	LONG i, j;
	
	if (g_hint || (number && g_mark_exclusion)) {
		rc.top = g_origin.y + y * g_grid_size;
		rc.bottom = rc.top + g_grid_size;
		rc.left = g_origin.x;
		rc.right = g_origin.x + g_matrix_size;
		InvalidateRect(hwnd, &rc, FALSE);
		rc.top = g_origin.y;
		rc.bottom = g_origin.y + g_matrix_size;
		rc.left = g_origin.x + x * g_grid_size;
		rc.right = rc.left + g_grid_size;
		InvalidateRect(hwnd, &rc, FALSE);
		for (i = 0; i < 9; i ++) {
			for (j = 0; j < 9; j ++) {
				if (g_playing_board.i[i][j] == g_playing_board.i[y][x]) {
					rc.top = g_origin.y + i * g_grid_size;
					rc.bottom = rc.top + g_grid_size;
					rc.left = g_origin.x + j * g_grid_size;
					rc.right = rc.left + g_grid_size;
					InvalidateRect(hwnd, &rc, FALSE);
				}
			}
		}
	} else {
		rc.top = g_origin.y + y * g_grid_size;
		rc.bottom = rc.top + g_grid_size;
		rc.left = g_origin.x + x * g_grid_size;
		rc.right = rc.left + g_grid_size;
		InvalidateRect(hwnd, &rc, FALSE);
	}
	UpdateWindow(hwnd);
}

void paint_matrix(HDC hdc, LONG grid, LONG matrix, LPPOINT org,
	HFONT hFont, HFONT hSmallFont, HPEN hPen, HPEN hBoldPen)
{
	static const TCHAR lpChar[][2] = {
		TEXT("0"), 
		TEXT("1"), TEXT("2"), TEXT("3"), TEXT("4"), TEXT("5"),
		TEXT("6"), TEXT("7"), TEXT("8"), TEXT("9"),
		TEXT("x"),
	};
	static const TCHAR lpAlpha[][2] = {
		TEXT("A"), TEXT("B"), TEXT("C"), TEXT("D"), TEXT("E"),
		TEXT("F"), TEXT("G"), TEXT("H"), TEXT("I"),
	};
	RECT rc;
	int i, j, k;
	unsigned int m;
	LONG g, gg, g2, x, y;
	HGDIOBJ obj[4] = {NULL};
	UINT uTypes[4] = {OBJ_BRUSH, OBJ_FONT, OBJ_PAL, OBJ_PEN, };
	
	get_current_objects(hdc, 4, uTypes, obj);
	
	g = 3 * grid / 11;
	gg = grid / 10;
	g2 = grid / 2;
	if (g < 1) g = 1;
	if (gg < 1) gg = 1;
	if (g2 < 1) gg = 1;
	if (g_paint_region) {
		for (i = 0; i < 9; i ++) {
			for (j = 0; j < 9; j ++) {
				k = g_board.i[i][j];
				if (g_hRegionBrush[i]) {
					rc.left = org->x + grid * j;
					rc.right = rc.left + grid;
					rc.top = org->y + grid * i;
					rc.bottom = rc.top + grid;
					FillRect(hdc, &rc, g_hRegionBrush[k]);
				}
			}
		}
	} else if (g_hRegionBrush[0]) {
		rc.left = org->x;
		rc.right = rc.left + matrix;
		rc.top = org->y;
		rc.bottom = rc.top + matrix;
		FillRect(hdc, &rc, g_hRegionBrush[0]);
	}
	SetBkMode(hdc, TRANSPARENT);
	/* axis */
	if (g_paint_axis) {
		if (hSmallFont) {
			SelectObject(hdc, hSmallFont);
		}
		for (i = 0; i < 9; i ++) {
			rc.top = org->y - g2;
			rc.bottom = org->y;
			rc.left = org->x + i * grid;
			rc.right = rc.left + grid;
			DrawText(hdc, lpAlpha[i], 1, &rc,
				DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
				DT_VCENTER);
			
			rc.top = org->y + matrix;
			rc.bottom = rc.top + g2;
			DrawText(hdc, lpAlpha[i], 1, &rc,
				DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
				DT_VCENTER);
			
			rc.top = org->y + i * grid;
			rc.bottom = rc.top + grid;
			rc.left = org->x - g2;
			rc.right = org->x;
			DrawText(hdc, lpChar[i + 1], 1, &rc,
				DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
				DT_VCENTER);
			
			rc.left = org->x + matrix;
			rc.right = rc.left + g2;
			DrawText(hdc, lpChar[i + 1], 1, &rc,
				DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
				DT_VCENTER);
		}
	}
	/* frame line */
	for (i = 0; i < 9; i ++) {
		if (g_playing_board.i[0][i] != g_playing_board.i[8][i]) {
			if (hBoldPen) {
				SelectObject(hdc, hBoldPen);
			}
		} else {
			if (hPen) {
				SelectObject(hdc, hPen);
			}
		}
		x = org->x + grid * i;
		MoveToEx(hdc, x, org->y, NULL);
		LineTo(hdc, x + grid, org->y);
		
		if (g_playing_board.i[i][0] != g_playing_board.i[i][8]) {
			if (hBoldPen) {
				SelectObject(hdc, hBoldPen);
			}
		} else {
			if (hPen) {
				SelectObject(hdc, hPen);
			}
		}
		y = org->y + grid * i;
		MoveToEx(hdc, org->x, y, NULL);
		LineTo(hdc, org->x, y + grid);
	}
	/* board */
	for (i = 0; i < 9; i ++) {
		for (j = 0; j < 9; j ++) {
			k = (i + 1) % 9;
			if (g_playing_board.i[i][j] != g_playing_board.i[k][j]) {
				if (hBoldPen) {
					SelectObject(hdc, hBoldPen);
				}
			} else {
				if (hPen) {
					SelectObject(hdc, hPen);
				}
			}
			x = org->x + grid * j;
			y = org->y + grid * i + grid;
			MoveToEx(hdc, x, y, NULL);
			LineTo(hdc, x + grid, y);
			
			k = (j + 1) % 9;
			if (g_playing_board.i[i][j] != g_playing_board.i[i][k]) {
				if (hBoldPen) {
					SelectObject(hdc, hBoldPen);
				}
			} else {
				if (hPen) {
					SelectObject(hdc, hPen);
				}
			}
			x = org->x + grid * j + grid;
			y = org->y + grid * i;
			MoveToEx(hdc, x, y, NULL);
			LineTo(hdc, x, y + grid);
			SelectObject(hdc, hPen);
			if (g_playing_board.n[i][j]) {
				/* number */
				if (hFont) {
					SelectObject(hdc, hFont);
				}
				rc.top = org->y + i * grid;
				rc.bottom = rc.top + grid;
				rc.left = org->x + j * grid;
				rc.right = rc.left + grid;
				if (g_board.n[i][j]) {
					SetTextColor(hdc, g_fix_color);
				} else {
					SetTextColor(hdc, g_number_color);
				}
				DrawText(hdc, lpChar[bit2int(g_playing_board.n[i][j])],
					1, &rc,
					DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
					DT_VCENTER);
			} else {
				/* hint/mark */
				if (hSmallFont) {
					SelectObject(hdc, hSmallFont);
				}
				SetTextColor(hdc, g_mark_color);
				if (g_hint) {
					m = ~(g_playing_board.x[j] | g_playing_board.y[i] |
					      g_playing_board.b[g_playing_board.i[i][j]]);
				} else if (g_mark_exclusion) {
					m = g_mark[i][j] & ~(g_playing_board.x[j] | g_playing_board.y[i] |
					    g_playing_board.b[g_playing_board.i[i][j]]);
				} else {
					m = g_mark[i][j];
				}
				m &= g_mark_mask;
				y = grid - gg;
				for (k = 0, m &= 0x1ff; m ; k ++, m >>= 1) {
					if (k % 3) {
						x += g;
					} else {
						x = gg; y -= g;
					}
					if (m & 0x1) {
						rc.top = org->y + i * grid + y;
						rc.bottom = rc.top + g;
						rc.left = org->x + j * grid + x;
						rc.right = rc.left + g;
						if (g_mark_type) {
							DrawText(hdc, lpChar[k + 1],
								1, &rc,
								DT_CENTER | DT_EXTERNALLEADING | DT_SINGLELINE |
								DT_VCENTER);
						} else {
							Ellipse(hdc, rc.left, rc.top, rc.right, rc.bottom);
						}
					}
				}
			}
		}
	}
	set_current_objects(hdc, 4, obj);
}

void on_mark_mode(HWND hwnd)
{
	g_mark_mode = ! g_mark_mode;
	update_menu_item(GetMenu(hwnd), IDM_MARK, g_mark_mode);
	SendMessage(g_hwndToolbar, TB_CHECKBUTTON, IDM_MARK,
	            MAKELONG(g_mark_mode, 0));
	
	update_status();
}

void on_edit_menu(HWND hwnd, int x, int y, int px, int py, BOOL mode)
{
	INT result;
	HMENU hMenu;
	
	if (x >= 0 && y >= 0 && ! g_board.n[y][x]) {
		hMenu = GetSubMenu((mode) ? g_hmenuContextMark : g_hmenuContext, 0);
		update_context_menu(hMenu, x, y);
		result = TrackPopupMenuEx(hMenu, 
			TPM_VCENTERALIGN | TPM_LEFTALIGN | TPM_NONOTIFY | TPM_RETURNCMD | TPM_RIGHTBUTTON,
			px, py, hwnd, NULL);
		/* process result */
		if (result > 0 && result <= 10) {
			if (! put_number(&g_playing_board, x, y, result - 1)) {
				update_status();
				update_matrix(hwnd, x, y, TRUE);
			}
		} else if (result >= 12 && result <= 21) {
			mark_toggle(x, y, result - 12);
			update_matrix(hwnd, x, y, FALSE);
		} else if (result == 22) {
			on_mark_mode(hwnd);
		} else if (result == 0) {
			/* eat WM_LBUTTONDOWN message */
			MSG msg;
			while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE)) {
				if (msg.message == WM_LBUTTONDOWN) {
					PeekMessage(&msg, NULL, 0, 0, PM_REMOVE);
				} else {
					break;
				}
			}
		}
	}
}

void on_click(HWND hwnd, int x, int y, BOOL menu)
{
	POINT pt;
	
	if (g_state == EDIT) {
		pt.x = x; pt.y = y;
		screen_to_matrix(&x, &y);
		if (x >= 0 && y >= 0) {
			if (! g_board.n[y][x]) {
				if (! menu || count_possible_numbers(&g_playing_board, x, y) < 2) {
					next(&g_playing_board, x, y);
					update_status();
					update_matrix(hwnd, x, y, TRUE);
				} else if (ClientToScreen(hwnd, &pt)) {
					on_edit_menu(hwnd, x, y, pt.x, pt.y, g_mark_mode);
				}
			}
		} else {
			on_mark_mode(hwnd);
		}
	}
}

void on_copy(HWND hwnd, Board *board, int size, void (*conv)(Board *, LPTSTR))
{
	LPTSTR str;
	
	str = malloc(sizeof(TCHAR) * size);
	if (str) {
		(*conv)(board, str);
		trace("%d %d", lstrlen(str), size);
		assert(lstrlen(str) < size);
		trace("%d", setclipboard(hwnd, str));
		free(str);
	}
}

void on_copy_metafile(HWND hwnd)
{
	LOGFONT logfont = {0};
	LONG size, grid;
	POINT origin;
	HFONT hFont, hSmallFont;
	HPEN hPen, hBoldPen;
	HDC hdc;
	HENHMETAFILE hmeta;
	RECT rc;
	
	grid = 60;
	if (g_paint_axis) {
		size = 596;
		origin.x = 24;
		origin.y = 24;
	} else {
		size = 556;
		origin.x = 4;
		origin.y = 4;
	}
	
	logfont.lfHeight = 48;
	logfont.lfWidth = 30;
#if defined(WINVER) && WINVER >= 0x0400
	logfont.lfQuality = ANTIALIASED_QUALITY;
#else
	logfont.lfQuality = DEFAULT_QUALITY;
#endif
	lstrcpy(logfont.lfFaceName, "Arial");
	hFont = CreateFontIndirect(&logfont);
	
	logfont.lfHeight = 19;
	logfont.lfWidth = 10;
	lstrcpy(logfont.lfFaceName, "Arial");
	hSmallFont = CreateFontIndirect(&logfont);
	
	hPen = CreatePen(PS_SOLID, 2, RGB(0, 0, 0));
	hBoldPen = CreatePen(PS_SOLID, 6, RGB(0, 0, 0));
	
	rc.top = rc.left = 0;
	rc.right = rc.bottom = size;
	hdc = create_window_enh_metafile(hwnd, NULL, &rc, NULL);
	if (hdc) {
		paint_matrix(hdc, grid, grid * 9, &origin, hFont, hSmallFont, hPen, hBoldPen);
		hmeta = CloseEnhMetaFile(hdc);
		if (hmeta) {
			if (setclipboard_handle(hwnd, hmeta, CF_ENHMETAFILE)) {
				DeleteEnhMetaFile(hmeta);
			}
		}
	}
	
	if (hSmallFont) {
		DeleteObject(hSmallFont);
	}
	if (hFont) {
		DeleteObject(hFont);
	}
	if (hPen) {
		DeleteObject(hPen);
	}
	if (hBoldPen) {
		DeleteObject(hBoldPen);
	}
}


void generate_thread_proc(void *param)
{
	HWND hwnd = (HWND) param;
	Board *b;
	int r, s = SYM_N;
	BOOL result = FALSE;
	
	trace("generate start");
	set_dlg_item_text_id(g_hInstance, hwnd, IDC_LABEL, IDS_GENERATING);
	b = (Board *) SetWindowLongPtr(hwnd, DWL_USER, 0);
	change_seed();
	zero_board(b);
	trace("generate region start");
	switch (g_symmetry) {
	case SYMMETRY_POLAR:
		s = SYM_P;
		break;
	case SYMMETRY_RANDOM:
		s = range_rand(SYM_MAX - 1) + 1;
		assert(s > 0 && s < SYM_MAX);
		break;
	}
	if (generate_region(b, g_region, s)) {
		trace("generate region end");
		trace("generate generate start");
		if (generate(b, -1, 0) == -1) {
			trace("generate generate success");
			if (g_symmetry == SYMMETRY_NONE) {
				r = random_scan_rm(b);
			} else {
				r = random_scan_rm_s(b, s);
			}
			if (!r && g_state == GENERATE) {
				result = TRUE;
			}
		}
	}
	EndDialog(hwnd, result);
	trace("generate end");
	_endthread();
}

void on_generate(HWND hwnd)
{
	if (g_state == EDIT) {
		Board zero;
		BOOL result;
		void *param[2];
		
		param[0] = generate_thread_proc;
		param[1] = &zero;
		g_state = GENERATE;
		result = DialogBoxParam(g_hInstance, "IDD_WAIT", hwnd,
			WaitDialogProc, (LPARAM) param);
		if (result) {
			g_board = g_playing_board = zero;
			mark_clear_all();
			update_copy_menu(GetMenu(hwnd), ! g_region);
			set_default_filename(hwnd);
			update_status();
			redraw(hwnd, NULL);
		}
	}
	g_state = EDIT;
}

void on_manipulate(HWND hwnd, Manipulate m)
{
	switch (m) {
	case NEW:
		initialize_board(&g_board);
		g_playing_board = g_board;
		mark_clear_all();
		update_copy_menu(GetMenu(hwnd), TRUE);
		set_default_filename(hwnd);
		break;
	case CLEAR:
		clear_number(&g_playing_board);
		clear_number(&g_board);
		mark_clear_all();
		set_default_filename(hwnd);
		break;
	case FIX:
		g_board = g_playing_board;
		break;
	case RETRY:
		g_playing_board = g_board;
		break;
	case RELEASE:
		clear_number(&g_board);
		break;
	}
	g_state = EDIT;
	update_status();
	redraw(hwnd, NULL);
}

void on_mark(HWND hwnd, int x, int y, int n)
{
	if (g_state == EDIT) {
		screen_to_matrix(&x, &y);
		if (x >= 0 && y >= 0) {
			mark_toggle(x, y, n);
			update_matrix(hwnd, x, y, FALSE);
		}
	}
}

void on_open_as(HWND hwnd)
{
	OPENFILENAME ofn = {0};
	TCHAR lpFile[256] = {0};
	
	g_state = EDIT;
#if defined(WINVER) && _WIN32_WINNT >= 0x0500
	ofn.lStructSize = OPENFILENAME_SIZE_VERSION_400;
#else
	ofn.lStructSize = sizeof(ofn);
#endif
	lstrcpy(lpFile, g_lpszProfileFileName);
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = _T("ini Files(*.ini)\0*.ini\0\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = lpFile;
	ofn.nMaxFile = 256;
	ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY |
	            OFN_NOCHANGEDIR;
	
	if (! GetOpenFileName(&ofn)) return;
	on_open(hwnd, lpFile);
}

void on_paste(HWND hwnd)
{
	Board zero;
	LPTSTR str;
	
	g_state = EDIT;
	str = getclipboard(hwnd);
	if (str) {
		if (ascii2board(&zero, str)) {
			g_board = g_playing_board = zero;
			mark_clear_all();
			update_copy_menu(GetMenu(hwnd), ! is_jigsaw(&g_board));
			set_default_filename(hwnd);
			update_status();
			redraw(hwnd, NULL);
		}
		free(str);
	}
}

void on_paste_region(HWND hwnd)
{
	Board zero;
	LPTSTR str;
	
	g_state = EDIT;
	str = getclipboard(hwnd);
	if (str) {
		if (ascii2region(&zero, str)) {
			g_board = g_playing_board = zero;
			mark_clear_all();
			update_copy_menu(GetMenu(hwnd), ! is_jigsaw(&g_board));
			set_default_filename(hwnd);
			update_status();
			redraw(hwnd, NULL);
		}
		free(str);
	}
}

void on_put(HWND hwnd, int x, int y, int n)
{
	if (g_state == EDIT) {
		screen_to_matrix(&x, &y);
		if (x >= 0 && y >= 0 && ! g_board.n[y][x]) {
			if (! put_number(&g_playing_board, x, y, n)) {
				update_status();
				update_matrix(hwnd, x, y, TRUE);
			}
		}
	}
}

void on_save_as(HWND hwnd)
{
	OPENFILENAME ofn = {0};
	SYSTEMTIME st;
	TCHAR lpFile[MAX_PATH] = {0};
	
	GetLocalTime(&st);
	wsprintf(lpFile, "%s-%02d.%02d-%02d.%02d.%02d.ini", g_lpszApplicationName,
		st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond);
	
#if defined(WINVER) && _WIN32_WINNT >= 0x0500
	ofn.lStructSize = OPENFILENAME_SIZE_VERSION_400;
#else
	ofn.lStructSize = sizeof(ofn);
#endif
	ofn.hwndOwner = hwnd;
	ofn.lpstrFilter = _T("ini Files(*.ini)\0*.ini\0\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFile = lpFile;
	ofn.nMaxFile = 256;
	ofn.Flags = OFN_CREATEPROMPT | OFN_OVERWRITEPROMPT |
	            OFN_EXPLORER | OFN_HIDEREADONLY | OFN_NOCHANGEDIR;
	
	if (! GetSaveFileName(&ofn)) return;
	on_save(hwnd, lpFile);
}


void solve_thread_proc(void *param)
{
	HWND hwnd = (HWND) param;
	Board *b;
	
	trace("solve start");
	b = (Board *) GetWindowLongPtr(hwnd, DWL_USER);
	set_dlg_item_text_id(g_hInstance, hwnd, IDC_LABEL, IDS_SOLVING);
	search(b, -1, 0);
	if (g_state == SOLVE) {
		switch (b->sol) {
		case 0:
			set_dlg_item_text_id(g_hInstance, hwnd, IDC_LABEL, IDS_IMPOSSIBLE);
			break;
		case 1:
			set_dlg_item_text_id(g_hInstance, hwnd, IDC_LABEL, IDS_POSSIBLE);
			break;
		default:
			set_dlg_item_text_id(g_hInstance, hwnd, IDC_LABEL, IDS_DUPLICITY);
			break;
		}
	}
	trace("solve end");
	_endthread();
}


void on_solve(HWND hwnd)
{
	if (g_state == EDIT) {
		Board zero;
		void *param[2];
		
		param[0] = solve_thread_proc;
		param[1] = &zero;
		g_state = SOLVE;
		zero = g_playing_board;
		zero.sol = 0;
		DialogBoxParam(g_hInstance, "IDD_WAIT", hwnd,
			WaitDialogProc, (LPARAM) param);
	}
	g_state = EDIT;
}

void OnClose(HWND hwnd)
{
	g_state = EDIT;
	if (g_autosave) {
		save_config(hwnd);
		on_save(hwnd, NULL);
	}
	DestroyWindow(hwnd);
}

void OnCommand(HWND hwnd, int id, HWND hwndCtl, UINT codeNotify)
{
	switch (id) {
	case IDM_OPEN:
		on_open_as(hwnd);
		break;
	case IDM_SAVE:
		on_save(hwnd, NULL);
		break;
	case IDM_SAVEAS:
		on_save_as(hwnd);
		break;
	case IDM_EXIT:
		g_state = EDIT;
		PostMessage(hwnd, WM_CLOSE, 0, 0);
		break;
	/* */
	case IDM_NEW:
		on_manipulate(hwnd, NEW);
		break;
	case IDM_CLEAR:
		on_manipulate(hwnd, CLEAR);
		break;
	case IDM_FIX:
		on_manipulate(hwnd, FIX);
		break;
	case IDM_RETRY:
		on_manipulate(hwnd, RETRY);
		break;
	case IDM_RELEASE:
		on_manipulate(hwnd, RELEASE);
		break;
	case IDM_GENERATE:
		on_generate(hwnd);
		break;
	case IDM_SOLVE:
		on_solve(hwnd);
		break;
	case IDM_CANCEL:
		g_state = EDIT;
		return;
	case IDM_MARK:
		on_mark_mode(hwnd);
		break;
	case IDM_MARK_EXCLUSION:
		g_mark_exclusion = ! g_mark_exclusion;
		update_menu_item(GetMenu(hwnd), IDM_MARK_EXCLUSION, g_mark_exclusion);
		if (! g_hint) {
			redraw(hwnd, NULL);
		}
		break;
	case IDM_MARK_ALL:
		mark_all(&g_playing_board);
		if (! g_hint) {
			redraw(hwnd, NULL);
		}
		break;
	case IDM_MARK_CLEAR:
		mark_clear_all();
		if (! g_hint) {
			redraw(hwnd, NULL);
		}
		break;
	/* copy */
	case IDM_COPY:
		on_copy(hwnd, &g_playing_board, B_ASCIILEN, board2ascii);
		break;
	case IDM_COPY_SPC:
		on_copy(hwnd, &g_playing_board, B_ASCIISPCLEN, board2ascii_spc);
		break;
	case IDM_COPY_ZUNTAN:
		on_copy(hwnd, &g_playing_board, B_ZUNTANLEN, board2zuntan);
		break;
	case IDM_COPY_ZUNTAN_URL:
		on_copy(hwnd, &g_playing_board, B_ZUNTAN_URL_LEN, board2zuntanurl);
		break;
	case IDM_COPY_BRAINBASHERS:
		on_copy(hwnd, &g_playing_board, B_BRAINBASHERS_LEN, board2brainbashers);
		break;
	case IDM_COPY_PLAYR_CO_UK:
		on_copy(hwnd, &g_playing_board, B_PLAYR_CO_UK_LEN, board2playr_co_uk);
		break;
	case IDM_COPY_CSV:
		on_copy(hwnd, &g_playing_board, B_CSVLEN, board2csv);
		break;
	case IDM_COPY_AAMB:
		on_copy(hwnd, &g_playing_board, B_AAMBLEN, board2aamb);
		break;
	case IDM_COPY_AAMB_FIXED:
		on_copy(hwnd, &g_playing_board, B_AAMBLEN, board2aamb_fixed);
		break;
	case IDM_COPY_AXIS:
		on_copy(hwnd, &g_playing_board, B_ASCIIAXISLEN, board2ascii_axis);
		break;
	case IDM_COPY_AXIS_FIXED:
		on_copy(hwnd, &g_playing_board, B_ASCIIAXISLEN, board2ascii_axis_fixed);
		break;
	/* */
	case IDM_COPY_PROBLEM:
		on_copy(hwnd, &g_board, B_ASCIILEN, board2ascii);
		break;
	case IDM_COPY_SPC_PROBLEM:
		on_copy(hwnd, &g_board, B_ASCIISPCLEN, board2ascii_spc);
		break;
	case IDM_COPY_ZUNTAN_PROBLEM:
		on_copy(hwnd, &g_board, B_ZUNTANLEN, board2zuntan);
		break;
	case IDM_COPY_ZUNTAN_URL_PROBLEM:
		on_copy(hwnd, &g_board, B_ZUNTAN_URL_LEN, board2zuntanurl);
		break;
	case IDM_COPY_BRAINBASHERS_PROBLEM:
		on_copy(hwnd, &g_board, B_BRAINBASHERS_LEN, board2brainbashers);
		break;
	case IDM_COPY_PLAYR_CO_UK_PROBLEM:
		on_copy(hwnd, &g_board, B_PLAYR_CO_UK_LEN, board2playr_co_uk);
		break;
	case IDM_COPY_CSV_PROBLEM:
		on_copy(hwnd, &g_board, B_CSVLEN, board2csv);
		break;
	case IDM_COPY_AAMB_PROBLEM:
		on_copy(hwnd, &g_board, B_AAMBLEN, board2aamb);
		break;
	case IDM_COPY_AAMB_FIXED_PROBLEM:
		on_copy(hwnd, &g_board, B_AAMBLEN, board2aamb_fixed);
		break;
	case IDM_COPY_AXIS_PROBLEM:
		on_copy(hwnd, &g_board, B_ASCIIAXISLEN, board2ascii_axis);
		break;
	case IDM_COPY_AXIS_FIXED_PROBLEM:
		on_copy(hwnd, &g_board, B_ASCIIAXISLEN, board2ascii_axis_fixed);
		break;
	case IDM_COPY_META:
		on_copy_metafile(hwnd);
		break;
	/* */
	case IDM_PASTE:
		on_paste(hwnd);
		break;
	case IDM_PASTE_REGION:
		on_paste_region(hwnd);
		break;
	/* view */
	case IDM_HINT:
		g_hint = ! g_hint;
		update_menu_item(GetMenu(hwnd), IDM_HINT, g_hint);
		redraw(hwnd, NULL);
		break;
	case IDM_HINT_ALL:
		change_mark_shown(hwnd, 10);
		break;
	case IDM_HINT_HIDE:
		change_mark_shown(hwnd, 11);
		break;
	case IDM_HINT_1:
		change_mark_shown(hwnd, 1);
		break;
	case IDM_HINT_2:
		change_mark_shown(hwnd, 2);
		break;
	case IDM_HINT_3:
		change_mark_shown(hwnd, 3);
		break;
	case IDM_HINT_4:
		change_mark_shown(hwnd, 4);
		break;
	case IDM_HINT_5:
		change_mark_shown(hwnd, 5);
		break;
	case IDM_HINT_6:
		change_mark_shown(hwnd, 6);
		break;
	case IDM_HINT_7:
		change_mark_shown(hwnd, 7);
		break;
	case IDM_HINT_8:
		change_mark_shown(hwnd, 8);
		break;
	case IDM_HINT_9:
		change_mark_shown(hwnd, 9);
		break;
	case IDM_AXIS:
		g_paint_axis = ! g_paint_axis;
		update_menu_item(GetMenu(hwnd), IDM_AXIS, g_paint_axis);
		redraw(hwnd, NULL);
		break;
	case IDM_COLOR:
		g_paint_region = ! g_paint_region;
		update_menu_item(GetMenu(hwnd), IDM_COLOR, g_paint_region);
		redraw(hwnd, NULL);
		break;
	/* config */
	case IDM_NO_SYMMETRY:
		change_symmetry_setting(hwnd, SYMMETRY_NONE);
		break;
	case IDM_SYMMETRY:
		change_symmetry_setting(hwnd, SYMMETRY_RANDOM);
		break;
	case IDM_SYMMETRY_POLAR:
		change_symmetry_setting(hwnd, SYMMETRY_POLAR);
		break;
	case IDM_JIGSAW:
		g_region = ! g_region;
		update_menu_item(GetMenu(hwnd), IDM_JIGSAW, g_region);
		break;
	case IDM_MARK_TYPE:
		g_mark_type = ! g_mark_type;
		update_menu_item(GetMenu(hwnd), IDM_MARK_TYPE, g_mark_type);
		redraw(hwnd, NULL);
		break;
	case IDM_AUTOSAVE:
		g_autosave = ! g_autosave;
		update_menu_item(GetMenu(hwnd), IDM_AUTOSAVE, g_autosave);
		break;
	case IDM_SAVECONFIG:
		if (! save_config(hwnd)) {
			message_box_id(g_hInstance, hwnd, IDS_SAVE_ERROR, g_lpszApplicationName, MB_ICONASTERISK | MB_OK);
		}
		break;
	/*
	case IDM_MINIMIZE:
		ShowWindow(hwnd, SW_MINIMIZE);
		break;
	*/
	}
}

void OnContextMenu(HWND hwnd, HWND hwndContext, UINT xPos, UINT yPos)
{
	POINT pt, ptc;
	int x, y;
	
	if (xPos == 0xffff || yPos == 0xffff) {
		if (! GetCursorPos(&pt)) {
			return;
		}
	} else {
		pt.x = xPos;  pt.y = yPos;
	}
	if (SendMessage(hwnd, WM_NCHITTEST, 0, MAKELPARAM(pt.x, pt.y)) != HTCLIENT) {
		FORWARD_WM_CONTEXTMENU(hwnd, hwndContext, pt.x, pt.y, DefWindowProc);
		return;
	}
	if (g_state != EDIT) return;
	ptc = pt;
	ScreenToClient(hwnd, &ptc);
	x = ptc.x;
	y = ptc.y;
	screen_to_matrix(&x, &y);
	on_edit_menu(hwnd, x, y, pt.x, pt.y, ! g_mark_mode);
}

BOOL OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
	TBBUTTON tbStdButtons1[] = {
		{STD_FILENEW, IDM_NEW, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{STD_PASTE, IDM_PASTE, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{STD_FILEOPEN, IDM_OPEN, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
	};
	TBBUTTON tbButtons1[] = {
		{13, IDM_GENERATE, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{0, 0, TBSTATE_ENABLED, BTNS_SEP, 0L, 0},
	};
	TBBUTTON tbStdButtons2[] = {
		{STD_FILESAVE, IDM_SAVE, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{0, 0, TBSTATE_ENABLED, BTNS_SEP, 0L, 0},
	};
	TBBUTTON tbButtons2[] = {
		{0, IDM_SOLVE, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{0, 0, TBSTATE_ENABLED, BTNS_SEP, 0L, 0},
		{10, IDM_MARK, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{0, 0, TBSTATE_ENABLED, BTNS_SEP, 0L, 0},
		{1, IDM_HINT_1, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{2, IDM_HINT_2, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{3, IDM_HINT_3, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{4, IDM_HINT_4, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{5, IDM_HINT_5, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{6, IDM_HINT_6, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{7, IDM_HINT_7, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{8, IDM_HINT_8, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{9, IDM_HINT_9, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{11, IDM_HINT_ALL, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
		{12, IDM_HINT_HIDE, TBSTATE_ENABLED, BTNS_BUTTON, 0L, 0},
	};
	TBADDBITMAP tbab;
	int i, n;
	
	g_hwndToolbar = CreateToolbarEx(
		hwnd,
		WS_CHILD | WS_VISIBLE | TBSTYLE_TOOLTIPS | CCS_TOP,
		ID_TOOLBAR, 0, HINST_COMMCTRL, IDB_STD_SMALL_COLOR,
		tbStdButtons1, sizeof(tbStdButtons1) / sizeof(tbStdButtons1[0]),
		0, 0, 0, 0, sizeof(TBBUTTON)
	);
	tbab.hInst = g_hInstance;
	tbab.nID = IDI_TOOLBAR;
	n = SendMessage(g_hwndToolbar, TB_ADDBITMAP, 13, (LPARAM) &tbab);
	for (i = 0; i < sizeof(tbButtons1) / sizeof(tbButtons1[0]); i++) {
		if (tbButtons1[i].idCommand != 0) {
			tbButtons1[i].iBitmap += n;
		}
	}
	for (i = 0; i < sizeof(tbButtons2) / sizeof(tbButtons2[0]); i++) {
		if (tbButtons2[i].idCommand != 0) {
			tbButtons2[i].iBitmap += n;
		}
	}
	Toolbar_TB_ADDBUTTONS(g_hwndToolbar, tbButtons1, SendMessage);
	Toolbar_TB_ADDBUTTONS(g_hwndToolbar, tbStdButtons2, SendMessage);
	Toolbar_TB_ADDBUTTONS(g_hwndToolbar, tbButtons2, SendMessage);
	
	g_hwndStatus = CreateStatusWindow(
		WS_CHILD | WS_VISIBLE | SBARS_SIZEGRIP | CCS_BOTTOM,
		"", hwnd, ID_STATUS
	);
	if (g_hwndToolbar == NULL || g_hwndStatus == NULL) {
		trace("create toolbar or statusbar");
		return FALSE;
	}
	
	update_menu_item(GetMenu(hwnd), IDM_JIGSAW, g_region);
	update_menu_item(GetMenu(hwnd), IDM_MARK_TYPE, g_mark_type);
	update_menu_item(GetMenu(hwnd), IDM_AUTOSAVE, g_autosave);
	change_mark_shown(hwnd, 0);
	
	update_symmetry_menu(GetMenu(hwnd), MF_CHECKED);
	update_status();
	
	DragAcceptFiles(hwnd, TRUE);
	return TRUE;
}

void OnDestroy(HWND hwnd)
{
	g_state = EDIT;
	PostQuitMessage(0);
}

void OnDropFiles(HWND hwnd, HDROP hdrop)
{
	LPTSTR lpFile;
	int size, result;
	
	g_state = EDIT;
	size = DragQueryFile(hdrop, 0, NULL, 0) + 1;
	if (size > 1) {
		lpFile = malloc(sizeof(TCHAR) * size);
		if (lpFile) {
			result = DragQueryFile(hdrop, 0, lpFile, size) + 1;
			if (result == size) {
				on_open(hwnd, lpFile);
			}
			free(lpFile);
		}
	}
	DragFinish(hdrop);
}

BOOL OnEraseBkgnd(HWND hwnd, HDC hdc)
{
	volatile BOOL result = TRUE;
	return result;
}

void OnKey(HWND hwnd, UINT vk, BOOL fDown, int cRepeat, UINT flags)
{
	static LPCTSTR mark_keys = "QWERTYUIOP";
	POINT pt;
	int i;
	
	if (! GetCursorPos(&pt) || ! ScreenToClient(hwnd, &pt)) {
		return;
	}
	if (vk >= '0' && vk <= '9') {
		if (g_mark_mode) {
			on_mark(hwnd, pt.x, pt.y, vk - '0');
		} else {
			on_put(hwnd, pt.x, pt.y, vk - '0');
		}
		return;
	} else if (vk >= VK_NUMPAD0 && vk <= VK_NUMPAD9) {
		if (g_mark_mode) {
			on_mark(hwnd, pt.x, pt.y, vk - VK_NUMPAD0);
		} else {
			on_put(hwnd, pt.x, pt.y, vk - VK_NUMPAD0);
		}
		return;
	} else if (vk == VK_SPACE || vk == VK_RETURN) {
		on_click(hwnd, pt.x, pt.y, FALSE);
		return;
	} else if (vk == VK_DELETE || vk == VK_BACK) {
		on_put(hwnd, pt.x, pt.y, 0);
		return;
	} else if (vk == VK_UP || vk == 'K') {
		if (! screen_to_matrix_cell(&pt.x, &pt.y)) {
			pt.y -= g_grid_size;
		}
	} else if (vk == VK_DOWN || vk == 'J') {
		if (! screen_to_matrix_cell(&pt.x, &pt.y)) {
			pt.y += g_grid_size;
		}
	} else if (vk == VK_LEFT || vk == 'H') {
		if (! screen_to_matrix_cell(&pt.x, &pt.y)) {
			pt.x -= g_grid_size;
		}
	} else if (vk == VK_RIGHT || vk == 'L') {
		if (! screen_to_matrix_cell(&pt.x, &pt.y)) {
			pt.x += g_grid_size;
		}
	} else if (vk == VK_PRIOR) {
		g_cursor_pos = (g_cursor_pos + 8) % 9;
		screen_to_matrix_cell(&pt.x, &pt.y);
	} else if (vk == VK_NEXT) {
		g_cursor_pos = (g_cursor_pos + 1) % 9;
		screen_to_matrix_cell(&pt.x, &pt.y);
	} else {
		for (i = 0; i < 10; i ++) {
			if (vk == mark_keys[i]) {
				if (g_mark_mode) {
					on_put(hwnd, pt.x, pt.y, i + 1);
				} else {
					on_mark(hwnd, pt.x, pt.y, i + 1);
				}
				break;
			}
		}
		return;
	}
	if (pt.x < g_origin.x) {
		pt.x = g_origin.x + g_matrix_size - (g_origin.x - pt.x) % g_grid_size;
	} else if (pt.x > g_origin.x + g_matrix_size) {
		pt.x = g_origin.x + (pt.x - g_origin.x - g_matrix_size) % g_grid_size;
	}
	if (pt.y < g_origin.y) {
		pt.y = g_origin.y + g_matrix_size - (g_origin.y - pt.y) % g_grid_size;
	} else if (pt.y > g_origin.y + g_matrix_size) {
		pt.y = g_origin.y + (pt.y - g_origin.y - g_matrix_size) % g_grid_size;
	}
	if (ClientToScreen(hwnd, &pt)) {
		SetCursorPos(pt.x, pt.y);
	}
}

void OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
	on_click(hwnd, x, y, TRUE);
}

int OnMouseActivate(HWND hwnd, HWND hwndTopLevel, UINT codeHitTest, UINT msg)
{
	return (codeHitTest == HTCLIENT) ? MA_ACTIVATEANDEAT : MA_ACTIVATE;
}

LRESULT OnMouseWheel(HWND hwnd, int x, int y, int zDelta, UINT fwKeys)
{
	POINT pt;
	pt.x = x;
	pt.y = y;
	if (ScreenToClient(hwnd, &pt)) {
		on_click(hwnd, pt.x, pt.y, FALSE);
	}
	return 0;
}

LRESULT OnNotify(HWND hwnd, int wParam, NMHDR *lParam)
{
	switch (((LPNMHDR) lParam)->code) {
	case TTN_GETDISPINFO:
		{
			static const UINT tooltips[][2] = {
				{IDM_NEW, IDS_TIPNEW},
				{IDM_PASTE, IDS_TIPPASTE},
				{IDM_OPEN, IDS_TIPOPEN},
				{IDM_GENERATE, IDS_TIPGENERATE},
				{IDM_SAVE, IDS_TIPSAVE},
				{IDM_SOLVE, IDS_TIPSOLVE},
				{IDM_MARK, IDS_TIPMARK},
				{IDM_HINT_1, IDS_TIPSHOWHINT},
				{IDM_HINT_2, IDS_TIPSHOWHINT},
				{IDM_HINT_3, IDS_TIPSHOWHINT},
				{IDM_HINT_4, IDS_TIPSHOWHINT},
				{IDM_HINT_5, IDS_TIPSHOWHINT},
				{IDM_HINT_6, IDS_TIPSHOWHINT},
				{IDM_HINT_7, IDS_TIPSHOWHINT},
				{IDM_HINT_8, IDS_TIPSHOWHINT},
				{IDM_HINT_9, IDS_TIPSHOWHINT},
				{IDM_HINT_ALL, IDS_TIPSHOWHINTALL},
				{IDM_HINT_HIDE, IDS_TIPHIDEHINTALL},
			};
			int i;
			LPTOOLTIPTEXT lpttt = (LPTOOLTIPTEXT)lParam;
			lpttt->hinst = g_hInstance;
			for (i = 0; i < sizeof(tooltips) / sizeof(tooltips[0]); i++) {
				if (lpttt->hdr.idFrom == tooltips[i][0]) {
					lpttt->lpszText = MAKEINTRESOURCE(tooltips[i][1]);
					break;
				}
			}
		}
		break;
	}
	return 0;
}

void OnPaint(HWND hwnd)
{
	PAINTSTRUCT paint;
	HDC hdc;
	HBITMAP hbm, hbm_org;
	HBRUSH hBrush;
	RECT rc = {0};
	
	if (! GetClientRect(hwnd, &rc)) return;
	BeginPaint(hwnd, &paint);
	hdc = CreateCompatibleDC(paint.hdc);
	if (hdc) {
		hbm = CreateCompatibleBitmap(paint.hdc, rc.right, rc.bottom);
		if (hbm) {
			hbm_org = (HBITMAP) SelectObject(hdc, hbm);
			hBrush = GetStockObject(WHITE_BRUSH);
			FillRect(hdc, &rc, hBrush);
			paint_matrix(hdc, g_grid_size, g_matrix_size, &g_origin, 
			             g_hFont, g_hSmallFont, g_hPen, g_hBoldPen);
			BitBlt(paint.hdc, 0, 0, rc.right, rc.bottom, hdc, 0, 0, SRCCOPY);
			SelectObject(hdc, hbm_org);
			DeleteObject(hBrush);
			DeleteObject(hbm);
		}
		DeleteDC(hdc);
	}
	EndPaint(hwnd, &paint);
}

void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
	RECT rcClient, rcToolbar, rcStatus;
	
	FORWARD_WM_SIZE(g_hwndToolbar, state, cx, cy, SendMessage);
	FORWARD_WM_SIZE(g_hwndStatus, state, cx, cy, SendMessage);
	if (! GetClientRect(hwnd, &rcClient) ||
	    ! GetWindowRect(g_hwndToolbar, &rcToolbar) ||
	    ! GetWindowRect(g_hwndStatus, &rcStatus)) {
		return;
	}
	rcClient.top += rcToolbar.bottom - rcToolbar.top;
	rcClient.bottom -= rcStatus.bottom - rcStatus.top;
	update_paint_param(&rcClient);
	redraw(hwnd, &rcClient);
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg) {
		HANDLE_MSG(hwnd, WM_CLOSE, OnClose);
		HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
		HANDLE_MSG(hwnd, WM_CONTEXTMENU, OnContextMenu);
		HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
		HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);
		HANDLE_MSG(hwnd, WM_DROPFILES, OnDropFiles);
//		HANDLE_MSG(hwnd, WM_ERASEBKGND, OnEraseBkgnd);
		HANDLE_MSG(hwnd, WM_KEYDOWN, OnKey);
		HANDLE_MSG(hwnd, WM_LBUTTONDOWN, OnLButtonDown);
		HANDLE_MSG(hwnd, WM_MOUSEWHEEL, OnMouseWheel);
//		HANDLE_MSG(hwnd, WM_MOUSEACTIVATE, OnMouseActivate);
		HANDLE_MSG(hwnd, WM_NOTIFY, OnNotify);
		HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
		HANDLE_MSG(hwnd, WM_SIZE, OnSize);
		default:
			return DefWindowProc(hwnd, uMsg, wParam, lParam);
	}
	return 0L;
}

BOOL init(HINSTANCE hInstance, HINSTANCE hPreInst, int nCmdShow)
{
	int x, y, cx, cy;
	WNDCLASS wc;
	
	g_hInstance = hInstance;
	
	g_hAccel = LoadAccelerators(hInstance, "IDA_ACCEL");
	if (g_hAccel == NULL) {
		return FALSE;
	}
	
	if (!hPreInst) {
		wc.style = 0;
		wc.lpfnWndProc = WndProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hInstance = hInstance;
		wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
		wc.hCursor = LoadCursor(NULL, IDC_ARROW);
		wc.hbrBackground = NULL;
		wc.lpszMenuName = "IDM_MENU";
		wc.lpszClassName = g_lpszApplicationName;
		if (!RegisterClass(&wc)) {
			return FALSE;
		}
	}
	load_config(&x, &y, &cx, &cy);
	g_hwndFrame = CreateWindowEx(0, g_lpszApplicationName,
		g_lpszApplicationName, WS_OVERLAPPEDWINDOW,
		x, y, cx, cy, NULL, NULL, g_hInstance, NULL);
	if (g_hwndFrame == NULL) {
		return FALSE;
	}
	g_hmenuContext = LoadMenu(hInstance, "IDM_CONTEXT");
	g_hmenuContextMark = LoadMenu(hInstance, "IDM_CONTEXT_MARK");
	if (g_hmenuContext == NULL ||
	    g_hmenuContextMark == NULL) {
		return FALSE;
	}
	set_default_filename(g_hwndFrame);
	if (g_autosave) {
		on_open(g_hwndFrame, g_lpszFileName);
	} else {
		initialize_board(&g_board);
		g_playing_board = g_board;
	}
	create_region_brush();
	ShowWindow(g_hwndFrame, nCmdShow);
	UpdateWindow(g_hwndFrame);
	return TRUE;
}

int is_canceled(void)
{
	return (g_state == EDIT);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPreInst,
	LPSTR lpszCmdLine, int nCmdShow)
{
	MSG msg;
	
	InitCommonControls();
	init_rand();
	if (init(hInstance, hPreInst, nCmdShow)) {
		while (GetMessage(&msg, NULL, 0, 0) > 0) {
			if (! TranslateAccelerator(g_hwndFrame, g_hAccel, &msg)) {
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
	}
	release_resource();
	return (msg.wParam);
}
